package NET.worlds.scape;

import NET.worlds.console.Console;
import NET.worlds.console.Cursor;
import NET.worlds.console.Gamma;
import NET.worlds.console.RenderCanvas;
import NET.worlds.console.Window;
import NET.worlds.network.URL;
import java.io.IOException;
import java.awt.CheckboxMenuItem;

public class Camera
  extends WObject
  implements MouseButtonHandler, MouseEnterHandler, MouseExitHandler, MouseMoveHandler, IncrementalRestorer, URLSelf
{
  private RenderCanvas canvas;
  static final int CAMERA_SIZE = 20;
  
  public Camera()
  {
    setVisible(false);
  }
  
  static
  {
    System.loadLibrary("capture_rwx");
    cameraNativeInit();
  }
  
  public Window getWindow()
  {
    if (this.canvas == null) {
      return null;
    }
    return this.canvas.getWindow();
  }
  
  private static boolean captureRequested() {
      return requestCapture != null && requestCapture.getState();
  }
  
  public BoundBoxTemp getBoundBox()
  {
    Transform t = getObjectToWorldMatrix();
    Point3Temp p = t.getPosition();
    t.recycle();
    
    Point3Temp q = Point3Temp.make(p).minus(20.0F);
    return BoundBoxTemp.make(q, p.plus(20.0F));
  }
  
  private boolean mouseIsOverClickable = false;
  private boolean validMouseCoordinates = false;
  private float validMouseX = -1.0F;
  private float validMouseY = -1.0F;
  private static WObject cachedWObject = null;
  private static float cachedMouseX = -1.0F;
  private static float cachedMouseY = -1.0F;
  private int cameraID;
  private int cameraMode;
  private boolean alwaysClearBackground;
  private float xPick;
  private float yPick;
  private Point3Temp unpickSpot;
  private WObject pickedObj;
  private static CheckboxMenuItem requestCapture;
  
  public static WObject getMousePickWObject()
  {
    return cachedWObject;
  }
  
  public static float getMousePickX()
  {
    return cachedMouseX;
  }
  
  public static float getMousePickY()
  {
    return cachedMouseY;
  }
  
  protected WObject updateView(Window w, int mode)
  {
    Room r = getRoom();
    if ((!hasClump()) && (r != null)) {
      r.aboutToDraw();
    }
    if(captureRequested() && r != null && r == Pilot.getActiveRoom()) {
      requestCapture.setState(false);
      int result = writeShape(r.clumpID, r.getEnvironment().clumpID, r.getInfiniteBackground().clumpID);
      NET.worlds.console.Console.println("Capture RWX: " + result);
    }
    Transform t = null;
    SuperRoot o = getOwner();
    
    Pilot pilot = null;
    if (((o instanceof Pilot)) && (o.getOwner() == r))
    {
      pilot = (Pilot)o;
      if (r != null)
      {
        pilot.aboutToDraw();
        
        detach();
        r.add(this);
        
        t = getTransform();
        
        Point3Temp motion = t.getPosition().times(pilot)
          .minus(pilot.getPosition());
        post(pilot);
        if (motion.squaredLength() > 1.0F)
        {
          moveBy(Point3Temp.make().minus(motion));
          
          boolean pBumpable = pilot.getBumpable();
          pilot.setBumpable(false);
          moveThrough(motion);
          pilot.setBumpable(pBumpable);
          
          Room r2 = getRoom();
          if ((!hasClump()) && (r2 != null) && (r2 != r)) {
            r2.aboutToDraw();
          }
        }
      }
    }
    WObject result = renderScene(w.getHwnd(), w.fullWidth(), 
      w.fullHeight(), mode, pilot);
    if (t != null)
    {
      detach();
      pilot.add(this);
      setTransform(t);
      t.recycle();
    }
    return result;
  }
  
  public void renderToCanvas()
  {
    if(requestCapture == null) {
      requestCapture = new CheckboxMenuItem("Capture RWX", false);
      Console.getActive().getMenu("Options").add(requestCapture);
    }
    if (!isActive()) {
      return;
    }
    Window w = getWindow();
    if (w == null) {
      return;
    }
    if (this.validMouseCoordinates)
    {
      this.xPick = this.validMouseX;
      this.yPick = this.validMouseY;
      
      WObject obj = updateView(w, 
        7);
      
      boolean toggleMouseCursor = false;
      
      cachedWObject = obj;
      cachedMouseX = this.validMouseX;
      cachedMouseY = this.validMouseY;
      if ((obj != null) && (obj.acceptsLeftClicks()))
      {
        if (!this.mouseIsOverClickable)
        {
          this.mouseIsOverClickable = true;
          toggleMouseCursor = true;
        }
      }
      else if (this.mouseIsOverClickable)
      {
        this.mouseIsOverClickable = false;
        toggleMouseCursor = true;
      }
      if (toggleMouseCursor)
      {
        URL newURL = null;
        if (this.mouseIsOverClickable) {
          newURL = URL.make("home:HAND-M.CUR");
        } else {
          newURL = URL.make("system:DEFAULT_CURSOR");
        }
        Console.getActive().getCursor().setURL(newURL);
      }
    }
    else
    {
      updateView(w, 6);
    }
  }
  
  public RenderCanvas getCanvas()
  {
    return this.canvas;
  }
  
  public void setCanvas(RenderCanvas c)
  {
    this.canvas = c;
  }
  
  public void setAlwaysClearBackground(boolean in)
  {
    this.alwaysClearBackground = in;
  }
  
  public Point3Temp lastPickSpot()
  {
    return this.unpickSpot;
  }
  
  public Transform lookAround = Transform.make();
  private static final int isBotMode = 0;
  private static final int isPickingMode = 1;
  private static final int isDrawingMode = 2;
  private static final int isThroughMode = 4;
  
  public void transferFrom(Camera target)
  {
    this.lookAround.setTransform(target.lookAround);
  }
  
  public WObject getObjectAt(float x, float y, boolean pickingThrough, Point3Temp pt)
  {
    Window w = getWindow();
    if (w == null) {
      return null;
    }
    this.xPick = x;
    this.yPick = y;
    this.unpickSpot = null;
    WObject obj = updateView(w, 1 + (
      pickingThrough ? 4 : 0));
    if ((pt != null) && (this.unpickSpot != null)) {
      pt.copy(this.unpickSpot);
    }
    return obj;
  }
  
  private void sendClick(WObject obj, MouseButtonEvent e)
  {
    WObject oldTarget = e.target;
    e.target = obj;
    obj.deliver(e);
    e.target = oldTarget;
  }
  
  static WObject[] downWentTo = new WObject[3];
  static Point3 downAt = new Point3();
  
  public boolean handle(MouseButtonEvent e)
  {
    WObject obj = null;
    
    int which = 3;
    if (e.key == 58113) {
      which = 0;
    } else if (e.key == 58114) {
      which = 1;
    } else if (e.key == 58116) {
      which = 2;
    }
    if ((e instanceof MouseUpEvent))
    {
      obj = downWentTo[which];
      downWentTo[which] = null;
    }
    else
    {
      Window w = getWindow();
      if ((w != null) && (!w.getDeltaMode()))
      {
        if ((which == 1) && (Gamma.getShaper() != null))
        {
          obj = getObjectAt(e.x, e.y, false, null);
          if (obj != null) {
            obj.rightMenu();
          }
        }
        else
        {
          obj = getObjectAt(e.x, e.y, true, downAt);
        }
        downWentTo[which] = obj;
      }
    }
    if (obj != null)
    {
      if ((which == 0) && ((e instanceof MouseUpEvent))) {
        obj.doAction(Console.getActive().getDefaultAction(), e);
      }
      sendClick(obj, e);
    }
    return false;
  }
  
  public boolean handle(MouseEnterEvent e)
  {
    if (e != null)
    {
      URL defaultCursor = URL.make("system:DEFAULT_CURSOR");
      Console.getActive().getCursor().setURL(defaultCursor);
      
      this.mouseIsOverClickable = false;
      this.validMouseCoordinates = true;
      this.validMouseX = e.x;
      this.validMouseY = e.y;
      
      cachedWObject = null;
      cachedMouseX = -1.0F;
      cachedMouseY = -1.0F;
    }
    return false;
  }
  
  public boolean handle(MouseExitEvent e)
  {
    if (e != null)
    {
      URL defaultCursor = URL.make("system:DEFAULT_CURSOR");
      Console.getActive().getCursor().setURL(defaultCursor);
      
      this.mouseIsOverClickable = false;
      this.validMouseCoordinates = false;
      this.validMouseX = -1.0F;
      this.validMouseY = -1.0F;
      
      cachedWObject = null;
      cachedMouseX = -1.0F;
      cachedMouseY = -1.0F;
    }
    return false;
  }
  
  public boolean handle(MouseMoveEvent e)
  {
    if (e != null)
    {
      this.validMouseX = e.x;
      this.validMouseY = e.y;
    }
    return false;
  }
  
  public Object properties(int index, int offset, int mode, Object value)
    throws NoSuchPropertyException
  {
    Object ret = null;
    
    ret = super.properties(index, offset + 0, mode, value);
    
    return ret;
  }
  
  private static Object classCookie = new Object();
  public float xForRoomPreAndPostRender;
  public float yForRoomPreAndPostRender;
  public float zForRoomPreAndPostRender;
  public float dForRoomPreAndPostRender;
  
  public void saveState(Saver s)
    throws IOException
  {
    s.saveVersion(2, classCookie);
    super.saveState(s);
  }
  
  public void restoreState(Restorer r)
    throws IOException, TooNewException
  {
    switch (r.restoreVersion(classCookie))
    {
    case 1: 
      super.restoreState(r);
      r.restoreFloat();
      r.restore();
      r.restore();
      break;
    case 0: 
    case 2: 
      super.restoreState(r);
      break;
    default: 
      throw new TooNewException();
    }
  }
  
  public int incRestore(int state, Restorer r, URLSelfLoader p)
    throws Exception
  {
    if (state == 0) {
      restoreState(r);
    }
    if (r.version() != 3) {
      return -1;
    }
    if (state == 0) {
      p.otemp1 = r.restore(false);
    }
    World w = (World)p.otemp1;
    if (state == 0) {
      w.setSourceURL(getSourceURL());
    }
    return w.incRestore(state, r, p);
  }
  
  public void incRef()
  {

  }
  
  public void decRef()
  {

  }
  
  public static native void cameraNativeInit();
  
  public native void nDrawText(String paramString, int paramInt1, int paramInt2, int paramInt3, int paramInt4);
  
  synchronized native WObject renderScene(int paramInt1, int paramInt2, int paramInt3, int paramInt4, Pilot paramPilot);
  
  private static native int writeShape(int clumpID, int envClumpID, int infBackClumpID);
}
